==> fetch.php?media=resources_multi_hardware:lorad_1.0.5.tar.gz: lorad_1.0.5/core_v1/clock.c <==
// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "clock.h"

#include <stdlib.h>
#include <sys/types.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "common/statistics.h"
#include "core.h"

static inline void debug_drift(const char *name, struct clock_sync *sync,
			       uint32_t counter, struct timespec *time)
{
	int32_t drift;

	if (sync->sync) {
		drift = (counter - sync->counter) - time_diff_us(&sync->time, time);
		debug("Demodulator synchronized with %s: drift=%ius", name, drift);

	} else {
		debug("Demodulator synchronized with %s", name);
	}
}

enum sync_status {
	SYNC_UNKNOWN,
	SYNC_DELAYED,
	SYNC_QUICK_RESYNC,
	SYNC_FAILED,
	SYNC_DONE,
};

static enum sync_status synchronize(struct core *core)
{
	struct clock_core *clock_core = core->clock->core;
	int counter_sampled_on_pps = get_counter_sampled_on_pps();
	int pps_enabled = core->clock->pps_enabled;
	struct clock_sync monotonic = {0};
	struct clock_sync utc = {0};
	struct clock_sync tai = {0};

	// Synchronize TAI + UTC
	if (pps_enabled &&
	    counter_sampled_on_pps == 1 &&
	    get_counter(&tai.counter) == 0 &&
	    clock_get_tai_time(core->clock->tai_offset, &tai.time) == 0) {

		// Synchronize only if we are in the middle of a second
		if (tai.time.tv_nsec < 100000000 ||
		    tai.time.tv_nsec > 900000000) {
			return SYNC_DELAYED;
		}

		tai.time.tv_nsec = 0;
		tai.sync = 1;

		debug_drift("TAI", &clock_core->tai, tai.counter, &tai.time);

		// Synchronize UTC from TAI
		if (clock_convert_tai_to_utc(core->clock->tai_offset, &tai.time, &utc.time) == 0) {
			utc.counter = tai.counter;
			utc.sync = 1;

			debug_drift("UTC", &clock_core->utc, tai.counter, &utc.time);
		}
	}

	// Synchronize MONOTONIC
	if (set_counter_sampled_on_pps(0) == 0 &&
	    get_counter(&monotonic.counter) == 0 &&
	    clock_get_monotonic_time(&monotonic.time) == 0) {
		monotonic.sync = 1;

		debug_drift("MONOTONIC", &clock_core->monotonic, monotonic.counter, &monotonic.time);
	}

	// Synchronize UTC
	if (!utc.sync &&
	    set_counter_sampled_on_pps(0) == 0 &&
	    get_counter(&utc.counter) == 0 &&
	    clock_get_utc_time(&utc.time) == 0) {
		utc.sync = 1;

		debug_drift("UTC", &clock_core->utc, utc.counter, &utc.time);
	}

	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return SYNC_FAILED;
	}
	clock_core->tai = tai;
	clock_core->utc = utc;
	clock_core->monotonic = monotonic;
	pthread_mutex_unlock(&core->clock->core->mutex);

	// Sample counter on PPS for the next synchronization if PPS is enabled.
	if (pps_enabled) {
		set_counter_sampled_on_pps(1);

		// Request a quick re-synchronization if PPS was just enabled.
		if (counter_sampled_on_pps == 0) {
			return SYNC_QUICK_RESYNC;
		}
	}

	return (tai.sync || monotonic.sync) ? SYNC_DONE : SYNC_FAILED;
}

static void *sync_thread(void *arg)
{
	struct core *core = arg;
	enum sync_status status;

	sleep_sec(5, core->clock->core->running);

	while (core->clock->core->running) {
		for (status = SYNC_UNKNOWN; core->started && status != SYNC_DONE;) {
			status = synchronize(core);

			if (status == SYNC_DELAYED) {
				debug("Synchronization delayed");
				sleep_ms(100);

			} else if (status == SYNC_QUICK_RESYNC) {
				debug("Quick re-synchronization requested");
				sleep_ms(1000);

			} else if (status == SYNC_FAILED) {
				warning("Failed to synchronize");

			} else if (status == SYNC_DONE) {
				if (core->clock->core->synchronized == 0) {
					api_send_event_system(core, PROTO__EVENT__SYSTEM__EVENT__SYNCHRONIZED);
					core->clock->core->synchronized = 1;
				}
			}
		}

		sleep_sec(15, core->clock->core->running);
	}

	pthread_exit(NULL);
}

struct clock_core *clock_core_init(void)
{
	struct clock_core *clock_core = NULL;

	if ((clock_core = zalloc(sizeof(*clock_core))) == NULL) {
		eerror("Failed to allocate clock core");
		return NULL;
	}

	if (pthread_mutex_init(&clock_core->mutex, NULL) != 0) {
		error("Failed to initialize clock core mutex");
		free(clock_core);
		return NULL;
	}

	return clock_core;
}

int clock_core_start(struct core *core)
{
	core->clock->core->running = 1;

	if (pthread_create(&core->clock->core->sync_thread, NULL, sync_thread, core) != 0) {
		error("Failed to start clock synchronization thread");
		clock_core_shutdown(core);
		return -1;
	}

	return 0;
}

int clock_core_join(struct core *core)
{
	if (pthread_join(core->clock->core->sync_thread, NULL) != 0) {
		error("Failed to join clock synchronization thread");
		return -1;
	}

	return 0;
}

void clock_core_shutdown(struct core *core)
{
	if (core && core->clock && core->clock->core) {
		core->clock->core->running = 0;
	}
}

void clock_core_free(struct core *core)
{
	if (core && core->clock) {
		free(core->clock->core);
	}
}

static int get_time_from_counter(struct core *core,
				 uint32_t *orig_counter, struct timespec *orig_tp,
				 uint32_t counter, struct timespec *tp)
{
	uint32_t udiff;

	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	udiff = counter - *orig_counter;
	*tp = *orig_tp;
	time_add_us(tp, udiff);

	pthread_mutex_unlock(&core->clock->core->mutex);

	return 0;
}

int clock_get_utc_time_from_counter(struct core *core, uint32_t counter, struct timespec *tp)
{
	if (!core->clock->core->utc.sync) {
		return -1;
	}

	return get_time_from_counter(core,
				     &core->clock->core->utc.counter,
				     &core->clock->core->utc.time,
				     counter, tp);
}

int clock_get_tai_time_from_counter(struct core *core, uint32_t counter, struct timespec *tp)
{
	if (!core->clock->pps_enabled ||
	    !core->clock->core->tai.sync) {
		return -1;
	}

	return get_time_from_counter(core,
				     &core->clock->core->tai.counter,
				     &core->clock->core->tai.time,
				     counter, tp);
}

int clock_get_gps_time_from_counter(struct core *core, uint32_t counter, struct timespec *tp)
{
	struct timespec tai_time;

	if (clock_get_tai_time_from_counter(core, counter, &tai_time) < 0) {
		return -1;
	}

	clock_convert_tai_to_gps(&tai_time, tp);

	return 0;
}

static int get_counter_from_time(struct core *core,
				 struct timespec *orig_tp, uint32_t *orig_counter,
				 struct timespec *tp, uint32_t *counter)
{
	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	*counter = *orig_counter + time_diff_us(orig_tp, tp);

	pthread_mutex_unlock(&core->clock->core->mutex);

	return 0;
}

int clock_get_current_counter(struct core *core, uint32_t *counter)
{
	struct timespec tp;

	if (core->clock->pps_enabled &&
	    core->clock->core->tai.sync &&
	    clock_get_tai_time(core->clock->tai_offset, &tp) == 0 &&
	    get_counter_from_time(core,
				  &core->clock->core->tai.time,
				  &core->clock->core->tai.counter,
				  &tp, counter) == 0) {
		return 0;
	}

	if (core->clock->core->monotonic.sync &&
	    clock_gettime(CLOCK_MONOTONIC, &tp) == 0 &&
	    get_counter_from_time(core,
				  &core->clock->core->monotonic.time,
				  &core->clock->core->monotonic.counter,
				  &tp, counter) == 0) {
		return 0;
	}

	return -1;
}

int clock_get_last_pps_counter(struct core *core, uint32_t *counter)
{
	struct timespec tp;

	if (core->clock->pps_enabled &&
	    core->clock->core->tai.sync &&
	    clock_get_tai_time(core->clock->tai_offset, &tp) == 0) {
		tp.tv_nsec = 0;

		if (get_counter_from_time(core,
					  &core->clock->core->tai.time,
					  &core->clock->core->tai.counter,
					  &tp, counter) == 0) {
			return 0;
		}
	}

	return -1;
}

int clock_get_counter_from_utc_time(struct core *core, struct timespec *tp, uint32_t *counter)
{
	if (!core->clock->core->utc.sync) {
		return -1;
	}

	return get_counter_from_time(core,
				     &core->clock->core->utc.time,
				     &core->clock->core->utc.counter,
				     tp, counter);
}

int clock_get_counter_from_tai_time(struct core *core, struct timespec *tp, uint32_t *counter)
{
	if (!core->clock->pps_enabled ||
	    !core->clock->core->tai.sync) {
		return -1;
	}

	return get_counter_from_time(core,
				     &core->clock->core->tai.time,
				     &core->clock->core->tai.counter,
				     tp, counter);
}

int clock_get_counter_from_gps_time(struct core *core, struct timespec *tp, uint32_t *counter)
{
	struct timespec tai_time;

	clock_convert_gps_to_tai(tp, &tai_time);

	if (clock_get_counter_from_tai_time(core, &tai_time, counter) < 0) {
		return -1;
	}

	return 0;
}

==> fetch.php?media=resources_multi_hardware:lorad_1.2.0.tar.gz: lorad_1.2.0/core_v1/clock.c <==
// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "clock.h"

#include <stdlib.h>
#include <sys/types.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"

#include "api.h"
#include "common/statistics.h"
#include "core.h"

static inline void debug_drift(const char *name, struct clock_sync *sync,
			       uint32_t counter, struct timespec *time)
{
	int32_t drift;

	if (sync->sync) {
		drift = (counter - sync->counter) - time_diff_us(&sync->time, time);
		sync->sum_drift += drift;
		debug("Demodulator synchronized with %s: drift=%ius sum=%ius",
		      name, drift, sync->sum_drift);

	} else {
		debug("Demodulator synchronized with %s", name);
	}
}

enum sync_status {
	SYNC_UNKNOWN,
	SYNC_DELAYED,
	SYNC_QUICK_RESYNC,
	SYNC_FAILED,
	SYNC_DONE,
};

static enum sync_status synchronize(struct core *core)
{
	struct clock_core *clock_core = core->clock->core;
	int counter_sampled_on_pps = get_counter_sampled_on_pps();
	int pps_enabled = core->clock->pps_enabled;
	struct clock_sync monotonic = {0};
	struct clock_sync utc = {0};
	struct clock_sync tai = {0};

	// Synchronize TAI + UTC
	if (pps_enabled &&
	    counter_sampled_on_pps == 1 &&
	    get_counter(&tai.counter) == 0 &&
	    clock_get_tai_time(core->clock->tai_offset, &tai.time) == 0) {

		// Synchronize only if we are in the middle of a second
		if (tai.time.tv_nsec < 100000000 ||
		    tai.time.tv_nsec > 900000000) {
			return SYNC_DELAYED;
		}

		tai.time.tv_nsec = 0;
		tai.sync = 1;

		debug_drift("TAI", &clock_core->tai, tai.counter, &tai.time);

		// Synchronize UTC from TAI
		if (clock_convert_tai_to_utc(core->clock->tai_offset, &tai.time, &utc.time) == 0) {
			utc.counter = tai.counter;
			utc.sync = 1;

			debug_drift("UTC", &clock_core->utc, tai.counter, &utc.time);
		}
	}

	// Synchronize MONOTONIC
	if (set_counter_sampled_on_pps(0) == 0 &&
	    get_counter(&monotonic.counter) == 0 &&
	    clock_get_monotonic_time(&monotonic.time) == 0) {
		monotonic.sync = 1;

		debug_drift("MONOTONIC", &clock_core->monotonic, monotonic.counter, &monotonic.time);
	}

	// Synchronize UTC
	if (!utc.sync &&
	    set_counter_sampled_on_pps(0) == 0 &&
	    get_counter(&utc.counter) == 0 &&
	    clock_get_utc_time(&utc.time) == 0) {
		utc.sync = 1;

		debug_drift("UTC", &clock_core->utc, utc.counter, &utc.time);
	}

	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return SYNC_FAILED;
	}
	clock_core->tai.sync = tai.sync;
	clock_core->tai.counter = tai.counter;
	clock_core->tai.time = tai.time;
	clock_core->utc.sync = utc.sync;
	clock_core->utc.counter = utc.counter;
	clock_core->utc.time = utc.time;
	clock_core->monotonic.sync = monotonic.sync;
	clock_core->monotonic.counter = monotonic.counter;
	clock_core->monotonic.time = monotonic.time;
	pthread_mutex_unlock(&core->clock->core->mutex);

	// Sample counter on PPS for the next synchronization if PPS is enabled.
	if (pps_enabled) {
		set_counter_sampled_on_pps(1);

		// Request a quick re-synchronization if PPS was just enabled.
		if (counter_sampled_on_pps == 0) {
			return SYNC_QUICK_RESYNC;
		}
	}

	return (tai.sync || monotonic.sync) ? SYNC_DONE : SYNC_FAILED;
}

static void *sync_thread(void *arg)
{
	struct core *core = arg;
	enum sync_status status;

	sleep_sec(5, core->clock->core->running);

	while (core->clock->core->running) {
		for (status = SYNC_UNKNOWN; core->started && status != SYNC_DONE;) {
			status = synchronize(core);

			if (status == SYNC_DELAYED) {
				debug("Synchronization delayed");
				sleep_ms(100);

			} else if (status == SYNC_QUICK_RESYNC) {
				debug("Quick re-synchronization requested");
				sleep_ms(1000);

			} else if (status == SYNC_FAILED) {
				warning("Failed to synchronize");

			} else if (status == SYNC_DONE) {
				if (core->clock->core->synchronized == 0) {
					api_send_event_system(core, PROTO__EVENT__SYSTEM__EVENT__SYNCHRONIZED);
					core->clock->core->synchronized = 1;
				}
			}
		}

		sleep_sec(15, core->clock->core->running);
	}

	pthread_exit(NULL);
}

struct clock_core *clock_core_init(void)
{
	struct clock_core *clock_core = NULL;

	if ((clock_core = zalloc(sizeof(*clock_core))) == NULL) {
		eerror("Failed to allocate clock core");
		return NULL;
	}

	if (pthread_mutex_init(&clock_core->mutex, NULL) != 0) {
		error("Failed to initialize clock core mutex");
		free(clock_core);
		return NULL;
	}

	return clock_core;
}

int clock_core_start(struct core *core)
{
	core->clock->core->running = 1;

	if (pthread_create(&core->clock->core->sync_thread, NULL, sync_thread, core) != 0) {
		error("Failed to start clock synchronization thread");
		clock_core_shutdown(core);
		return -1;
	}

	return 0;
}

int clock_core_join(struct core *core)
{
	if (pthread_join(core->clock->core->sync_thread, NULL) != 0) {
		error("Failed to join clock synchronization thread");
		return -1;
	}

	return 0;
}

void clock_core_shutdown(struct core *core)
{
	if (core && core->clock && core->clock->core) {
		core->clock->core->running = 0;
	}
}

void clock_core_free(struct core *core)
{
	if (core && core->clock) {
		free(core->clock->core);
	}
}

static int get_time_from_counter(struct core *core,
				 uint32_t *orig_counter, struct timespec *orig_tp,
				 uint32_t counter, struct timespec *tp)
{
	uint32_t udiff;

	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	udiff = counter - *orig_counter;
	*tp = *orig_tp;
	time_add_us(tp, udiff);

	pthread_mutex_unlock(&core->clock->core->mutex);

	return 0;
}

int clock_get_utc_time_from_counter(struct core *core, uint32_t counter, struct timespec *tp)
{
	if (!core->clock->core->utc.sync) {
		return -1;
	}

	return get_time_from_counter(core,
				     &core->clock->core->utc.counter,
				     &core->clock->core->utc.time,
				     counter, tp);
}

int clock_get_tai_time_from_counter(struct core *core, uint32_t counter, struct timespec *tp)
{
	if (!core->clock->pps_enabled ||
	    !core->clock->core->tai.sync) {
		return -1;
	}

	return get_time_from_counter(core,
				     &core->clock->core->tai.counter,
				     &core->clock->core->tai.time,
				     counter, tp);
}

int clock_get_gps_time_from_counter(struct core *core, uint32_t counter, struct timespec *tp)
{
	struct timespec tai_time;

	if (clock_get_tai_time_from_counter(core, counter, &tai_time) < 0) {
		return -1;
	}

	clock_convert_tai_to_gps(&tai_time, tp);

	return 0;
}

static int get_counter_from_time(struct core *core,
				 struct timespec *orig_tp, uint32_t *orig_counter,
				 struct timespec *tp, uint32_t *counter)
{
	if (pthread_mutex_lock(&core->clock->core->mutex) != 0) {
		return -1;
	}

	*counter = *orig_counter + time_diff_us(orig_tp, tp);

	pthread_mutex_unlock(&core->clock->core->mutex);

	return 0;
}

int clock_get_current_counter(struct core *core, uint32_t *counter)
{
	struct timespec tp;

	if (core->clock->pps_enabled &&
	    core->clock->core->tai.sync &&
	    clock_get_tai_time(core->clock->tai_offset, &tp) == 0 &&
	    get_counter_from_time(core,
				  &core->clock->core->tai.time,
				  &core->clock->core->tai.counter,
				  &tp, counter) == 0) {
		return 0;
	}

	if (core->clock->core->monotonic.sync &&
	    clock_gettime(CLOCK_MONOTONIC, &tp) == 0 &&
	    get_counter_from_time(core,
				  &core->clock->core->monotonic.time,
				  &core->clock->core->monotonic.counter,
				  &tp, counter) == 0) {
		return 0;
	}

	return -1;
}

int clock_get_last_pps_counter(struct core *core, uint32_t *counter)
{
	struct timespec tp;

	if (core->clock->pps_enabled &&
	    core->clock->core->tai.sync &&
	    clock_get_tai_time(core->clock->tai_offset, &tp) == 0) {
		tp.tv_nsec = 0;

		if (get_counter_from_time(core,
					  &core->clock->core->tai.time,
					  &core->clock->core->tai.counter,
					  &tp, counter) == 0) {
			return 0;
		}
	}

	return -1;
}

int clock_get_counter_from_utc_time(struct core *core, struct timespec *tp, uint32_t *counter)
{
	if (!core->clock->core->utc.sync) {
		return -1;
	}

	return get_counter_from_time(core,
				     &core->clock->core->utc.time,
				     &core->clock->core->utc.counter,
				     tp, counter);
}

int clock_get_counter_from_tai_time(struct core *core, struct timespec *tp, uint32_t *counter)
{
	if (!core->clock->pps_enabled ||
	    !core->clock->core->tai.sync) {
		return -1;
	}

	return get_counter_from_time(core,
				     &core->clock->core->tai.time,
				     &core->clock->core->tai.counter,
				     tp, counter);
}

int clock_get_counter_from_gps_time(struct core *core, struct timespec *tp, uint32_t *counter)
{
	struct timespec tai_time;

	clock_convert_gps_to_tai(tp, &tai_time);

	if (clock_get_counter_from_tai_time(core, &tai_time, counter) < 0) {
		return -1;
	}

	return 0;
}
